/*
* DCF77 decoding
* (c) 27.6.2015 Ingo D. Rullhusen
*/

#include <limits.h>
#include <byteswap.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>

#include <alsa/asoundlib.h>


#include "DCF.h"


typedef short S16;

char *pcm_name = "default:CARD=DCsampler";
const snd_pcm_format_t format = SND_PCM_FORMAT_S16_LE;
const unsigned int channels = 2;
unsigned int rate = 192000;

snd_pcm_t *handle;

int open_mode = 0;
snd_pcm_stream_t stream = SND_PCM_STREAM_CAPTURE;
int nonblock = 0;

size_t bits_per_sample, bits_per_frame;
u_char *audiobuf = NULL;

S16 *sigdata1 = NULL;

int abuffsize = 1024;
int sigdatasize = 1024 / ( 2 * 2 );  // char stereo

int err;
int set_time_mode = 0;

FILE *Fpre;
FILE *Fpost;


void init_alsa()
{
   int err;

   snd_pcm_info_t *info;

   snd_pcm_info_alloca(&info);

   err = snd_pcm_open(&handle, pcm_name, stream, open_mode);
   if (err < 0)
   {
      fprintf(stderr, "audio open error: %s\n", snd_strerror(err));
      exit(1);
   }

   if ((err = snd_pcm_info(handle, info)) < 0)
   {
      fprintf(stderr, "info error: %s\n", snd_strerror(err));
      exit(1);
   }

   err = snd_pcm_nonblock(handle, nonblock);
   if (err < 0)
   {
      fprintf(stderr, "nonblock setting error: %s\n", snd_strerror(err));
      exit(1);
   }


   snd_pcm_hw_params_t *params;
   snd_pcm_hw_params_alloca(&params);
   err = snd_pcm_hw_params_any(handle, params);
   if (err < 0)
   {
      fprintf(stderr, "broken configuration for this PCM: no configurations available\n" );
      exit(1);
   }

/*****
   snd_pcm_format_t format;
   fprintf(stderr, "Available formats:\n");
   for (format = 0; format <= SND_PCM_FORMAT_LAST; format++)
   {
      if (snd_pcm_hw_params_test_format(handle, params, format) == 0)
         fprintf(stderr, "- %s\n", snd_pcm_format_name(format));
   }
*****/

   err = snd_pcm_hw_params_set_access(handle, params,
                                      SND_PCM_ACCESS_RW_INTERLEAVED);
   if (err < 0)
   {
      fprintf(stderr, "access type not available\n" );
      exit(1);
   }

   err = snd_pcm_hw_params_set_format(handle, params, format);
   if (err < 0)
   {
      fprintf(stderr, "sample format non available\n" );
      exit(1);
   }
   err = snd_pcm_hw_params_set_channels(handle, params, channels);
   if (err < 0)
   {
      fprintf(stderr, "channels count non available\n" );
      exit(1);
   }

   err = snd_pcm_hw_params_set_rate_near(handle, params, &rate, 0);
   if (err < 0)
   {
      fprintf(stderr, "rate not available\n" );
      exit(1);
   }

   err = snd_pcm_hw_params(handle, params);
   if (err < 0)
   {
      fprintf(stderr, "unable to install hw params\n" );
      exit(1);
   }


   bits_per_sample = snd_pcm_format_physical_width(format);
   bits_per_frame = bits_per_sample * channels;


}

void close_alsa()
{
   if (handle) snd_pcm_close(handle);
   handle = NULL;
}

int clean_exit()
{
   close_alsa();

///   fclose(Fpre);
///   fclose(Fpost);

   free(audiobuf);
   free(sigdata1);

   return(0);
}

static void signal_handler(int sig)
{

   if (sig == SIGABRT || sig==SIGTERM || sig==SIGINT)
   {
      fprintf(stderr, "aborted by signal %s...\n", strsignal(sig));

      exit( clean_exit() );
   }

   signal(sig, signal_handler);
}


void preprocess( S16 *data, int datasize )
{
   int i, k;

   const int TPno = 6;
   const double TPz[] = {
      0.0021390, -0.0051658, 0.0030863,
      0.0030863, -0.0051658, 0.0021390
                        };
   const double TPn[] = {
       0., -2.9000539, 2.8050405,  // 0. = 1
       -0.9048675, 0., 0.
                        };
   static double TPbuf[] = {0,0,0,0,0,0};


   const int Dno = 5;
   const double Dz[] = {
      -1.0/12.0, +8.0/12.0, 0.0, -8.0/12.0, +1.0/12.0
                       };
   const Dn[] = {
      0.0, 0, 0, 0, 0
                };
   static double Dbuf[] = {0,0,0,0,0};


   for ( i=0; i<datasize; ++i )
   {
      double X = abs( data[i] ) / (double)SHRT_MAX;

// Tiefpaß
      for ( k=0; k<TPno-1; ++k )
      { TPbuf[k] = X*TPz[k] + TPbuf[k+1] - TPbuf[0]*TPn[k]; }
      TPbuf[TPno-1] = X*TPz[k] - TPbuf[0]*TPn[k];

      X = TPbuf[0];

      data[i] = SHRT_MAX * X;
   }

}


#define AVGBLKS 3000

double level( S16 *data, int datasize )
{
   int i;
   double avg;

   static int next = 0;
   static double avg0[AVGBLKS];

   avg = 0;
   for ( i=0; i<datasize; ++i )
   {
      avg += data[i] / (double)SHRT_MAX;
   }
   avg /= (double)datasize;

   avg0[ next ] = avg;
   ++next;
   if ( next >= AVGBLKS ) { next = 0; }

   avg = 0;
   for ( i=0; i<AVGBLKS; ++i )
   {
      avg += avg0[i];
   }
   avg /= (double)AVGBLKS;

   return( avg );
}


void get_impulse( S16 *data, int datasize, double lvl0, double lvl1 )
{
   int i, iset;
   static int bit = 0;
   static double X0 = 0;

   iset = 0;
   for ( i=0; i<datasize; ++i )
   {
      double X = data[i] / (double)SHRT_MAX;

      if ( bit==1 && X<X0 && X < lvl0 )
      {
         bit = 0;
         if ( to0_dcf( i-iset ) && set_time_mode && dcf_data.ERROR==0 )
         {
            struct timeval tv;
         
            struct tm TIME;
            TIME.tm_sec = 0;
            TIME.tm_min = dcf_data.Min;
            TIME.tm_hour = dcf_data.Hr;
            TIME.tm_mday = dcf_data.Day;
            TIME.tm_mon = dcf_data.Month - 1;  // since january
            TIME.tm_year = 100 + dcf_data.Yr;  // since 1900
            TIME.tm_wday = -1;
            TIME.tm_yday = -1;
            TIME.tm_isdst = dcf_data.EST;

            tv.tv_sec = mktime( &TIME );
            tv.tv_usec = 0;

            if ( settimeofday(&tv,NULL) != 0 )
            {
               fprintf(stderr, "no permission to set clock\n" );
               clean_exit();
               exit(1);
            }

            err = 1;
         }
         iset = i;
      }
      if ( bit==0 && X>X0 && X > lvl1 )
      {
         bit = 1;
         to1_dcf( i-iset );
         iset = i;
      }

      X0 = X;

      data[i] = ( bit ? SHRT_MAX : 0 );
   }
   nochange_dcf( i-iset );

}


int main( int argc, char * const argv[] )
{
   int chidx, opt;
   double lvl;

   int channel = 1;
   long int frnomax = 0;

   while ((opt = getopt(argc, argv, "c::d::m::sh")) != -1)
   {
      switch (opt)
      {
         case 'c':
            channel = atoi(optarg);
            break;
         case 'd':
            pcm_name = optarg;
            break;
         case 'm':
            frnomax = atoi(optarg) * 60 * (192000 / 256);
            break;
         case 's':
            set_time_mode = 1;
            break;
         case 'h':
         default:
            fprintf(stderr, "Usage: %s [-h] [-d\"<ALSA device>\"] [-c<channel>] [-m<min>] [-s]\n",
                    argv[0]);
            exit(EXIT_FAILURE);
      }
   }

   printf("ALSA device: %s\n",pcm_name);
   switch ( channel )
   {
      case 1:
         printf("channel    : 1 (left)\n");
         chidx = 0;
         break;
      case 2:
         printf("channel    : 2 (right)\n");
         chidx = 1;
         break;
      default:
         fprintf(stderr, "wrong channel number\n" );
         exit(EXIT_FAILURE);
   }
   fflush(stdout);

   reset_dcf();

   init_alsa();

   signal(SIGINT, signal_handler);
   signal(SIGTERM, signal_handler);
   signal(SIGABRT, signal_handler);

   audiobuf = (u_char *)malloc(abuffsize);
   sigdata1 = (S16*)calloc( sigdatasize, sizeof(S16) );

///   Fpre  = fopen( "/tmp/tst1.raw", "w" );
///   Fpost = fopen( "/tmp/tst2.raw", "w" );

   snd_pcm_status_t *status;
   snd_pcm_status_alloca(&status);

   err = 0;

   int i;
   long int frno=0;
   do
   {
      size_t f =  abuffsize  * 8 / bits_per_frame;
      size_t fc = snd_pcm_readi( handle, audiobuf, f );

      snd_pcm_status(handle, status);
      if (snd_pcm_status_get_state(status) == SND_PCM_STATE_XRUN)
      {
         fprintf(stderr, "XRUN\n" );

// restart
         if ( (err = snd_pcm_prepare(handle)) < 0 )
         {
            fprintf(stderr, "xrun: prepare error: %s\n", snd_strerror(err) );
            clean_exit();
            return(1);
         }
      }

      if ( f == fc )
      {
///         fwrite( audiobuf, 1, abuffsize, Fpre );

         for ( i=0; i<sigdatasize; ++i )
         { sigdata1[i] = ((S16*)audiobuf)[chidx+2*i]; }

         preprocess( sigdata1, sigdatasize );

         lvl = level( sigdata1, sigdatasize );

         get_impulse( sigdata1, sigdatasize, lvl*0.75, lvl*0.8 );

///         fwrite( sigdata1, 2, sigdatasize, Fpost );

         ++frno;
      }

   } while( (frnomax>0 && frno<frnomax) || err==0 );

   return( clean_exit() );
}

